function ClickableSourceCode() {
    this.node = document.createElement('ol');
    this.node.className = 'clickablesourcecode';
    this.clear();
}

/** Function to call on change selected line request */
ClickableSourceCode.prototype.onchange = function(
    newIndex    // Requested line number to select
) {
    return true; // true → allow, false → disallow
}

/** Line numbers of not selectable lines */
ClickableSourceCode.prototype.notSelectable = [];

/** Present source code */
Object.defineProperty(ClickableSourceCode.prototype, 'code', { enumerable: true,
    set(code) {
        this.clear();
        this.appendCode(code);
    },
    get() {
        return [...this.node.getElementsByTagName('code')].map(node => node.textContent).join('\n');
    }
});

/** Highlighted line */
Object.defineProperty(ClickableSourceCode.prototype, 'line', { enumerable: true,
    set(linenumber) {
        const lines = [...this.node.childNodes];
        const end = lines.pop();
        lines.forEach((el, index) => {
            if (index == linenumber)
                el.classList.add('selected');
            else
                el.classList.remove('selected');
        });
        if (Number.isFinite(linenumber))
            end.classList.remove('selected');
        else
            end.classList.add('selected');
    }
});

/**
* Requests a change of the selected line and checks the request using this.onchange
* @return undefined
*/
ClickableSourceCode.prototype.requestSelectedChange = function(
    newIndex    // Requested line number to select
) {
    if (!(this.onchange instanceof Function) || this.onchange.call(this, newIndex))
        this.line = newIndex;
}

/**
* Get number of lines of present source code
* @return Number        Number of lines
*/
ClickableSourceCode.prototype.countLines = function() {
    return this.node.childNodes.length - 1;
}

/**
* Append source code
* @return undefined
*/
ClickableSourceCode.prototype.appendCode = function(
    code        // Java code to append
) {
    const preCount = this.countLines();
    code.split('\n').forEach((s, index) => {
        const nodeLi = this._addLi();
        if (this.notSelectable.indexOf(preCount + index) == -1)
            this._addBtn(nodeLi, preCount + index);
        else {
            const doc = nodeLi.ownerDocument;
            const abbr = doc.createElement('abbr');
            abbr.setAttribute('title', 'Nicht auswählbar');
            abbr.appendChild(doc.createTextNode('🚫'));
            nodeLi.appendChild(abbr);
        }
        this._addCode(nodeLi, s);
    });
}

/**
* Remove all lines from displayed source code
* @return undefined
*/
ClickableSourceCode.prototype.clear = function() {
    this.node.innerHTML = '';
    this._addBtn(this._addLi(), Infinity);
}

/**
* Add a list item at the end of the present source code
* @return HTMLListItemElement       Created DOM element
*/
ClickableSourceCode.prototype._addLi = function() {
    const doc = this.node.ownerDocument;
    const node = doc.createElement('li');
    this.node.insertBefore(node, this.node.lastChild);
    return node;
}

/**
* Adds a button to select this line of code to a list item
* @return undefined
*/
ClickableSourceCode.prototype._addBtn = function(
    nodeLi,         // HTMLElement to add button to
    linenumber      // Line number represent by button (may be Infinity for end of code)
) {
    if (!Number.isFinite(linenumber))
        linenumber = Infinity;
    const doc = this.node.ownerDocument;
    const btn = doc.createElement('button');
    btn.textContent = Number.isFinite(linenumber) ? 'Weiter mit:' : 'Fertig';
    btn.addEventListener('click', () => { this.requestSelectedChange(linenumber); });
    const span = doc.createElement('span');
    span.appendChild(btn);
    span.appendChild(doc.createTextNode(' '));
    nodeLi.appendChild(span);
}

/**
* Add formated Java code to list item
* @return undefined
*/
ClickableSourceCode.prototype._addCode = function(
    nodeLi,     // HTMLElement to add button to
    code        // Java code
) {
    const doc = this.node.ownerDocument;
    const nodePre = doc.createElement('pre');
    const nodeCode = doc.createElement('code');
    nodePre.appendChild(nodeCode);
    nodeCode.className = 'hljs language-java';
    nodeCode.innerHTML = hljs.highlight(code, {language: 'java'}).value;
    nodeLi.appendChild(nodePre);
}
